﻿#region Using directives
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using Telerik.Windows.Controls;
using Telerik.Windows.Data;
using System.ComponentModel;
using System.Linq.Expressions;
using System.Reflection; 
#endregion

namespace BC.DataFilterExtensions
{
    /// <summary>
    /// Extensions for RadDataFilter
    /// </summary>
    public static class Extensions
    {
        /// <summary>
        /// Serialize CompositeFilterDescriptorCollection to XML
        /// </summary>
        /// <param name="collection">CompositeFilterDescriptorCollection</param>
        /// <typeparam name="T">The type that the filters are based on</typeparam>
        /// <returns>The XML</returns>
        public static XDocument GetXDocument<T>(this CompositeFilterDescriptorCollection collection)
        {
            // Create XML document
            return new XDocument(new XDeclaration("1.0", null, null),
                                    // Create root element
                                    new XElement("FilterDescriptors",
                                        // Serialize type information
                                        new XElement("Type", typeof(T).FullName),
                                        new XElement("TypeAssembly", typeof(T).Assembly.GetName().Name),
                                        // Serialize the collection
                                        collection.GetXElement()));
        }

        /// <summary>
        /// Serialize IFilterDescriptor
        /// </summary>
        /// <param name="filter">FilterDescriptor</param>
        /// <returns>The XML</returns>
        public static XElement GetXElement(this IFilterDescriptor filter)
        {
            // Create return value
            XElement retVal = new XElement("FilterDescriptor", 
                                                // Serialize type information of current filter
                                                new XAttribute("FilterType", filter.GetType().FullName), 
                                                new XAttribute("FilterTypeAssembly", filter.GetType().Assembly.GetName().Name));

            // Is the parameter a FilterDescriptor instance?
            if (filter.GetType() == typeof(FilterDescriptor))
            {
                // Cast in FilterDescriptor
                FilterDescriptor filterDescriptor = (FilterDescriptor)filter;

                if (filterDescriptor.Value == Telerik.Windows.Data.FilterDescriptor.UnsetValue)
                    return null;

                // Serialize members
                retVal.Add(new XElement("IsCaseSensitive", filterDescriptor.IsCaseSensitive));
                retVal.Add(new XElement("Member", filterDescriptor.Member));
                retVal.Add(new XElement("MemberType", filterDescriptor.MemberType.FullName));
                retVal.Add(new XElement("MemberTypeAssembly", filterDescriptor.MemberType.Assembly.GetName().Name));
                retVal.Add(new XElement("Operator", filterDescriptor.Operator));
                retVal.Add(new XElement("Value", Convert.ToString(filterDescriptor.Value)));
            }
            // Does the filter instance implements ICompositeFilterDescriptor interface
            else if (typeof(ICompositeFilterDescriptor).IsAssignableFrom(filter.GetType()))
            {
                // Cast instance in ICompositeFilterDescriptor
                ICompositeFilterDescriptor compositeFilterDescriptor = (ICompositeFilterDescriptor)filter;

                // Serialize LogicalOperator
                retVal.Add(new XElement("LogicalOperator", compositeFilterDescriptor.LogicalOperator));

                // Serialize containing FilterDescriptors recursive
                retVal.Add(new XElement("FilterDescriptors", (from f in compositeFilterDescriptor.FilterDescriptors
                                                              select f.GetXElement()).Where(i => i != null)));
            }
            else
                // Other types are not implemented
                throw new NotImplementedException();

            return retVal;
        }

        /// <summary>
        /// Get FilterDescriptorCollection from XML
        /// </summary>
        /// <param name="document">XML document</param>
        /// <returns>CompositeFilterDescriptorCollection</returns>
        public static CompositeFilterDescriptorCollection GetFilterDescriptorCollection(this XDocument document)
        {
            // If no valid xml document is given ...
            if (document == null || document.Root == null || document.Root.Name != "FilterDescriptors")
                // return empty collection
                return new CompositeFilterDescriptorCollection();

            // Deserialize collection from XML
            return (CompositeFilterDescriptorCollection)document.Root.Element("FilterDescriptor").GetIFilterDescriptor();
        }

        /// <summary>
        /// Get IFilterDescriptor from XML
        /// </summary>
        /// <param name="element">XML element</param>
        /// <returns>IFilterDescriptor instance</returns>
        public static IFilterDescriptor GetIFilterDescriptor(this XElement element)
        {
            // Elementname must be "FilterDescriptor"
            if (element.Name != "FilterDescriptor")
                return null;

            // Get assembly and type
            Assembly assembly = AppDomain.CurrentDomain.GetAssemblies().First(a => a.GetName().Name == element.Attribute("FilterTypeAssembly").Value);
            Type filterType = assembly.GetType(element.Attribute("FilterType").Value);

            // Create instance from type
            IFilterDescriptor instance = (IFilterDescriptor)Activator.CreateInstance(filterType);

            // Is it an FilterDescriptor?
            if (instance is FilterDescriptor)
            {
                // Cast in FilterDescriptor
                FilterDescriptor filterDescriptor = (FilterDescriptor)instance;

                // Deserialize members
                filterDescriptor.IsCaseSensitive = bool.Parse(element.Element("IsCaseSensitive").Value);
                filterDescriptor.Member = element.Element("Member").Value;
                Assembly memberTypeAssembly = AppDomain.CurrentDomain.GetAssemblies().First(a => a.GetName().Name == element.Element("MemberTypeAssembly").Value);
                filterDescriptor.MemberType = memberTypeAssembly.GetType(element.Element("MemberType").Value);
                filterDescriptor.Operator = (FilterOperator)Enum.Parse(typeof(FilterOperator), element.Element("Operator").Value);

                var valueString = element.Element("Value").Value;

                // Deserialize filter value
                if (string.IsNullOrEmpty(valueString) == false)
                {
                    TypeConverter converter = TypeDescriptor.GetConverter(filterDescriptor.MemberType);
                    filterDescriptor.Value = converter.ConvertFromString(valueString);
                }
            }
            // Does Filter implement 
            else if (instance is ICompositeFilterDescriptor)
            {
                // Case in ICompositeFilterDescriptor
                ICompositeFilterDescriptor compositeFilterDescriptor = (ICompositeFilterDescriptor)instance;

                // Deserialize LogicalOperator
                compositeFilterDescriptor.LogicalOperator = (FilterCompositionLogicalOperator)Enum.Parse(typeof(FilterCompositionLogicalOperator), element.Element("LogicalOperator").Value);
                
                // Deserialize filter descriptors recursive
                compositeFilterDescriptor.FilterDescriptors.AddRange(from x in element.Element("FilterDescriptors").Elements("FilterDescriptor")
                                                                     select x.GetIFilterDescriptor());
            }

            return instance;
        }

        /// <summary>
        /// Set FilterDescriptors in RadDataFilter fom XML document
        /// </summary>
        /// <param name="radDataFilter">RadDataFilter instance</param>
        /// <param name="document">XML document</param>
        public static void SetFilterDescriptors(this RadDataFilter radDataFilter, XDocument document)
        {
            // Deserialize CompositeFilterDescriptorCollection
            CompositeFilterDescriptorCollection collection = document.GetFilterDescriptorCollection();

            // First clear all filters on RadDataFilter
            radDataFilter.FilterDescriptors.Clear();

            // Set LogicalOperator
            radDataFilter.FilterDescriptors.LogicalOperator = collection.LogicalOperator;

            // Set filters
            radDataFilter.FilterDescriptors.AddRange(collection);
        }

        /// <summary>
        /// Get filter method from FilterDescriptor
        /// </summary>
        /// <typeparam name="T">The type that the filter is based on</typeparam>
        /// <param name="filterDescriptor">FilterDescriptor</param>
        /// <returns>Dynamic compiled filter method</returns>
        public static Func<T, bool> GetFilter<T>(this IFilterDescriptor filterDescriptor)
        {
            // Create parameter of Type T
            ParameterExpression parameter = Expression.Parameter(typeof(T));

            // Create filter expression
            var result = filterDescriptor.CreateFilterExpression(parameter);

            // Compile Lambda expression
            return Expression.Lambda<Func<T, bool>>(result, parameter).Compile();
        }

        /// <summary>
        /// Does the filter match?
        /// </summary>
        /// <typeparam name="T">The type that the filter is based on</typeparam>
        /// <param name="item">Filtered item of type T</param>
        /// <param name="filterDescriptor">FilterDescriptor</param>
        /// <returns>Does the filter match?</returns>
        public static bool MatchesFilter<T>(this T item, IFilterDescriptor filterDescriptor)
        {
            // Get filter method
            var filter = filterDescriptor.GetFilter<T>();

            // Execute filter method with given parameter
            return filter(item);
        }

        /// <summary>
        /// Filter IEnumerable of type T
        /// </summary>
        /// <typeparam name="T">The type the filter is based on</typeparam>
        /// <param name="items">Items of type T</param>
        /// <param name="filterDescriptor">FilterDescriptor</param>
        /// <returns>Get the matched items</returns>
        public static IEnumerable<T> Filter<T>(this IEnumerable<T> items, IFilterDescriptor filterDescriptor)
        {
            // Get filter method
            var filter = filterDescriptor.GetFilter<T>();

            // Filter the items
            return from i in items
                   where i.MatchesFilter<T>(filterDescriptor)
                   select i;
        }
    }
}
