﻿#region Using directives
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Telerik.Windows.Controls;
using Telerik.Windows.Data;
using System.ComponentModel;
using System.Windows;
using System.Collections;
using System.Xml.Linq; 
#endregion

namespace BC.DataFilterExtensions
{
    /// <summary>
    /// Helper class for RadDataFilter
    /// </summary>
    public class UnboundRadDataFilterHelper<T> : DependencyObject, IDisposable
    {
        #region Declaration

        /// <summary>
        /// Is instance disposed?
        /// </summary>
        private bool _isDisposed = false;

        /// <summary>
        /// Is filter currently updating?
        /// </summary>
        private bool _isUpdatingFilter = false;

        /// <summary>
        /// Is filter xml currently changing?
        /// </summary>
        private bool _isFilterXmlChanging = false;

        #endregion

        #region Construction

        /// <summary>
        /// Initializes a new instance of the <see cref="UnboundRadDataFilterHelper"/> class.
        /// </summary>
        /// <param name="radDataFilter">RadDataFilter instance</param>
        public UnboundRadDataFilterHelper(RadDataFilter radDataFilter)
        {
            // Parameter must be given!
            if (radDataFilter == null)
                throw new ArgumentNullException("radDataFilter");

            // Set RadDataFilter
            this.RadDataFilter = radDataFilter;

            // Create EnumEditorTemplateSelector instance
            this.RadDataFilter.EditorTemplateSelector = new EnumEditorTemplateSelector();

            // Track source changes
            this.RadDataFilter.SourceUpdated += new EventHandler<System.Windows.Data.DataTransferEventArgs>(RadDataFilter_SourceUpdated);

            // Track editor creation
            this.RadDataFilter.EditorCreated += new EventHandler<Telerik.Windows.Controls.Data.DataFilter.EditorCreatedEventArgs>(RadDataFilter_EditorCreated);

            // Check Source property on RadDataFilter
            this.CheckSource();
        }

        #endregion

        #region Events

        #endregion

        #region Properties

        /// <summary>
        /// RadDataFilter instance
        /// </summary>
        public RadDataFilter RadDataFilter { get; private set; }

        /// <summary>
        /// Unfiltered elements
        /// </summary>
        public IEnumerable<T> Source
        {
            get { return (IEnumerable<T>)GetValue(SourceProperty); }
            set { SetValue(SourceProperty, value); }
        }

        // Using a DependencyProperty as the backing store for Source.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty SourceProperty =
            DependencyProperty.Register("Source", typeof(IEnumerable<T>), typeof(UnboundRadDataFilterHelper<T>), new UIPropertyMetadata(null));

        /// <summary>
        /// Filtered Source
        /// </summary>
        public IEnumerable<T> FilteredSource
        {
            get { return (IEnumerable<T>)GetValue(FilteredSourceProperty); }
            set { SetValue(FilteredSourceProperty, value); }
        }

        // Using a DependencyProperty as the backing store for FilteredSource.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty FilteredSourceProperty =
            DependencyProperty.Register("FilteredSource", typeof(IEnumerable<T>), typeof(UnboundRadDataFilterHelper<T>), new UIPropertyMetadata(null));

        /// <summary>
        /// Filter als XML
        /// </summary>
        public string FilterXml
        {
            get { return (string)GetValue(FilterXmlProperty); }
            set { SetValue(FilterXmlProperty, value); }
        }

        // Using a DependencyProperty as the backing store for FilterXml.  This enables animation, styling, binding, etc...
        public static readonly DependencyProperty FilterXmlProperty =
            DependencyProperty.Register("FilterXml", typeof(string), typeof(UnboundRadDataFilterHelper<T>), new UIPropertyMetadata(null));

        #endregion

        #region Methods

        /// <summary>
        /// Check source on RadDataFilter
        /// </summary>
        private void CheckSource()
        {
            // Only attach in Unbound mode
            if (this.RadDataFilter.Source == null)
            {
                // Attach events on FilterDescriptors
                this.AttachEvents(this.RadDataFilter.FilterDescriptors);

                // Update filter
                this.UpdateFilter();
            }
            else
            {
                throw new NotImplementedException("Source must be null!");
            }
        }

        /// <summary>
        /// Attach events on item recursive
        /// </summary>
        /// <param name="item">Item</param>
        private void AttachEvents(object item)
        {
            // Is instance disposed?
            if (this._isDisposed)
                throw new ObjectDisposedException("RadDataFilterHelper");

            // Implements item ICompositeFilterDescriptor?
            if (item is ICompositeFilterDescriptor)
            {
                // Watch collection changed events
                ((ICompositeFilterDescriptor)item).FilterDescriptors.CollectionChanged += new System.Collections.Specialized.NotifyCollectionChangedEventHandler(FilterDescriptors_CollectionChanged);

                // Attach events on child FilterDescriptors recursive
                foreach (var child in ((ICompositeFilterDescriptor)item).FilterDescriptors)
                    this.AttachEvents(child);
            }

            // Implements item INotifyPropertyChanged?
            if (item is INotifyPropertyChanged)
                // Watch property changes on filter Source
                ((INotifyPropertyChanged)item).PropertyChanged += new PropertyChangedEventHandler(FilterDescriptor_PropertyChanged);
        }

        /// <summary>
        /// Detach events on item recursive
        /// </summary>
        /// <param name="item">Item</param>
        private void DetachEvents(object item)
        {
            // Is instance disposed?
            if (this._isDisposed)
                throw new ObjectDisposedException("RadDataFilterHelper");

            // Implements item ICompositeFilterDescriptor?
            if (item is ICompositeFilterDescriptor)
            {
                // Don't watch collection changed events anymore
                ((ICompositeFilterDescriptor)item).FilterDescriptors.CollectionChanged -= new System.Collections.Specialized.NotifyCollectionChangedEventHandler(FilterDescriptors_CollectionChanged);

                // Detach events on child FilterDescriptors recursive
                foreach (var child in ((ICompositeFilterDescriptor)item).FilterDescriptors)
                    this.DetachEvents(child);
            }

            // Implements item INotifyPropertyChanged?
            if (item is INotifyPropertyChanged)
                // Don't watch property changes on filter Source anymore
                ((INotifyPropertyChanged)item).PropertyChanged -= new PropertyChangedEventHandler(FilterDescriptor_PropertyChanged);
        }

        /// <summary>
        /// Update filter
        /// </summary>
        private void UpdateFilter()
        {
            if (this._isFilterXmlChanging || this.Source == null)
                return;

            // Avoid Re-Entrance
            this._isUpdatingFilter = true;

            // Filter elements
            this.FilteredSource = this.Source.Filter(this.RadDataFilter.FilterDescriptors);

            // Update the XML
            this.FilterXml = this.RadDataFilter.FilterDescriptors.GetXDocument<T>().ToString();

            this._isUpdatingFilter = false;
        }

        /// <summary>
        /// Returns a <see cref="System.String"/> that represents this instance.
        /// </summary>
        /// <returns>
        /// A <see cref="System.String"/> that represents this instance.
        /// </returns>
        public override string ToString()
        {
            // Is instance disposed?
            if (this._isDisposed)
                return "Instance is disposed!";
            else
                return this.RadDataFilter.ToString();
        }

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            if (this.RadDataFilter != null)
            {
                this.DetachEvents(this.RadDataFilter.FilterDescriptors);

                // Detach SourceUpdated event
                this.RadDataFilter.SourceUpdated -= new EventHandler<System.Windows.Data.DataTransferEventArgs>(RadDataFilter_SourceUpdated);

                // Detach EditorCreated event
                this.RadDataFilter.EditorCreated += new EventHandler<Telerik.Windows.Controls.Data.DataFilter.EditorCreatedEventArgs>(RadDataFilter_EditorCreated);

                this.RadDataFilter = null;
            }

            this._isDisposed = true;
        }

        /// <summary>
        /// Invoked whenever the effective value of any dependency property on this <see cref="T:System.Windows.DependencyObject"/> has been updated. The specific dependency property that changed is reported in the event data.
        /// </summary>
        /// <param name="e">Event data that will contain the dependency property identifier of interest, the property metadata for the type, and old and new values.</param>
        protected override void OnPropertyChanged(DependencyPropertyChangedEventArgs e)
        {
            base.OnPropertyChanged(e);

            switch (e.Property.Name)
            {
                case "FilterXml":

                    // Avoid Re-Entrance
                    if (this._isUpdatingFilter)
                        return;

                    try
                    {
                        this._isFilterXmlChanging = true;

                        // Is value empty?
                        if (string.IsNullOrEmpty((string)e.NewValue))
                        {
                            // Then clear filter descriptors
                            this.RadDataFilter.FilterDescriptors.Clear();
                        }
                        else
                        {
                            // Parse XDocument
                            XDocument document = XDocument.Parse((string)e.NewValue);

                            // Set filter descriptors from xml document
                            this.RadDataFilter.SetFilterDescriptors(document);
                        }
                    }
                    catch (Exception ex)
                    {
                        // Set old value
                        this.FilterXml = e.OldValue as string;

                        // Show error message
                        MessageBox.Show(ex.Message, "Error", MessageBoxButton.OK, MessageBoxImage.Error);
                    }
                    finally
                    {
                        this._isFilterXmlChanging = false;
                    }

                    // Update filter
                    this.UpdateFilter();

                    break;

                case "Source":
                    this.UpdateFilter();
                    break;
            }
        }

        #endregion

        #region Event handlers

        /// <summary>
        /// Handles the EditorCreated event of the RadDataFilter control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="Telerik.Windows.Controls.Data.DataFilter.EditorCreatedEventArgs"/> instance containing the event data.</param>
        void RadDataFilter_EditorCreated(object sender, Telerik.Windows.Controls.Data.DataFilter.EditorCreatedEventArgs e)
        {
            if (e.ItemPropertyInfo.PropertyType.IsEnum && e.Editor is RadComboBox)
                // Set enum members as ItemsSource for ComboBox
                ((RadComboBox)e.Editor).ItemsSource = Enum.GetValues(e.ItemPropertyInfo.PropertyType);
        }

        /// <summary>
        /// Handles the CollectionChanged event of the FilterDescriptors control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Collections.Specialized.NotifyCollectionChangedEventArgs"/> instance containing the event data.</param>
        void FilterDescriptors_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                // Filter added
                case System.Collections.Specialized.NotifyCollectionChangedAction.Add:
                    foreach (var item in e.NewItems)
                        // Attach events on all added Filters
                        this.AttachEvents(item);
                    break;

                // Filter removed
                case System.Collections.Specialized.NotifyCollectionChangedAction.Remove:
                    foreach (var item in e.OldItems)
                        // Detach events on all removed filters
                        this.DetachEvents(item);
                    break;
            }

            this.UpdateFilter();
        }

        /// <summary>
        /// Handles the PropertyChanged event of the FilterDescriptor control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.ComponentModel.PropertyChangedEventArgs"/> instance containing the event data.</param>
        void FilterDescriptor_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            this.UpdateFilter();
        }

        /// <summary>
        /// Handles the SourceUpdated event of the RadDataFilter control.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.Windows.Data.DataTransferEventArgs"/> instance containing the event data.</param>
        void RadDataFilter_SourceUpdated(object sender, System.Windows.Data.DataTransferEventArgs e)
        {
            this.CheckSource();
        }

        #endregion
    }
}
